# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from logging import Logger as logger_t
from operator import attrgetter as GetAttribute
from operator import itemgetter as ItemAt
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import json_any.json_any as jsnr

import cell_tracking_BC.standard.issue as isse
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.cell.cell import cell_t
from cell_tracking_BC.type.track.single.forking import forking_track_t
from cell_tracking_BC.type.track.single.thread import thread_track_t
from cell_tracking_BC.type.track.single.structured import (
    cells_w_optional_time_points_h,
    feature_filtering_h,
    per_thread_cell_h,
    per_thread_cells_h,
    per_thread_feature_h,
    structured_track_t,
)
from cell_tracking_BC.type.track.single.unstructured import (
    unstructured_track_t,
    TIME_POINT,
)
from cell_tracking_BC.type.track.multiple.unstructured import unstructured_tracks_t


any_track_h = Union[unstructured_track_t, structured_track_t]

TrackIssues_h = Callable[
    [structured_track_t, dict], Optional[Union[str, Sequence[str]]]
]


@dtcl.dataclass(repr=False, eq=False)
class tracks_t(List[structured_track_t]):

    invalids: List[Tuple[any_track_h, Sequence[str]]] = dtcl.field(
        init=False, default_factory=list
    )
    fully_pruned: List[structured_track_t] = dtcl.field(
        init=False, default_factory=list
    )

    NewFromJsonString: ClassVar[Callable[[str], tracks_t]] = jsnr.ObjectFromJsonString

    @classmethod
    def NewFromUnstructuredTracks(cls, tracks: unstructured_tracks_t, /) -> tracks_t:
        """"""
        instance = cls()

        next_thread_label = 1
        for unstructured in tracks.track_iterator:
            cells = _TreeIncompatibleCells(unstructured)
            if cells is None:
                (
                    forking_track,
                    next_thread_label,
                ) = forking_track_t.NewFromUnstructuredTrack(
                    unstructured, next_thread_label
                )

                if forking_track.leaves.__len__() > 1:
                    instance.append(forking_track)
                else:
                    thread_track = forking_track.AsThreadTrack()
                    instance.append(thread_track)
            else:
                instance.invalids.append((unstructured, cells))

        return instance

    @classmethod
    def __NewFromJsonDescription__(
        cls,
        description,
        /,
    ) -> tracks_t:
        """"""
        tracks, attributes = description

        instance = cls.__new__(
            cls
        )  # Do not use cls(tracks) (reason? because it is a dataclass?)
        instance.extend(tracks)
        for key, value in attributes.items():
            setattr(instance, key, value)

        return instance

    def __DescriptionForJSON__(self) -> Any:
        """"""
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose information.
        attributes = {_fld.name: getattr(self, _fld.name) for _fld in dtcl.fields(self)}

        return list(self), attributes

    def FilterOut(
        self,
        TrackIssues: TrackIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        TrackIssues: Arguments are: track and (optional) keyword arguments; Returned value can be None, an str, or a
            sequence of str.
        kwargs: Passed to TrackIsInvalid as keyword arguments
        """
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            issues = TrackIssues(track, **kwargs)
            if issues is None:
                t_idx += 1
            else:
                self.invalids.append((track, issues))
                del self[t_idx]

    def RootCells(
        self, /, *, with_time_point: bool = False
    ) -> cells_w_optional_time_points_h:
        """"""
        if with_time_point:
            what = GetAttribute("root", "root_time_point")
        else:
            what = GetAttribute("root")

        return tuple(map(what, self))

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """"""
        for track in self:
            if isinstance(track, forking_track_t):
                track.MarkDividingCells(
                    before_deaths,
                    division_responses=division_responses,
                    lower_bound=lower_bound,
                    CombinedResponse=CombinedResponse,
                )

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_thread: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_thread_cells_h]:
        """"""
        if per_thread:
            output = {}
            AddDividing = output.update
        else:
            output = []
            AddDividing = output.extend

        for track in self:
            if isinstance(track, forking_track_t):
                dividing = track.DividingCells(
                    with_time_point=with_time_point, per_thread=per_thread
                )
                AddDividing(dividing)

        if with_time_point and not per_thread:
            output.sort(key=ItemAt(1))

        return output

    def DivisionTimePoints(self) -> Dict[int, Optional[Sequence[int]]]:
        """"""
        output = {}

        for track in self:
            if isinstance(track, thread_track_t):
                output[track.label] = None
            else:
                output.update(track.DivisionTimePoints())

        return output

    def MarkDeadCells(
        self,
        death_responses: Dict[int, Optional[Sequence[float]]],
        lower_bound: float,
        after_divisions: bool,
        /,
        *,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """
        A death is set if the highest response at a cell among the adjacent threads is above the threshold.
        """
        t_idx = 0
        while t_idx < self.__len__():
            track = self[t_idx]
            fully_pruned = track.MarkDeadCells(
                death_responses,
                lower_bound,
                after_divisions,
                CombinedResponse=CombinedResponse,
                called_from_educated_code=True,
            )
            if fully_pruned:
                self.fully_pruned.append(track)
                del self[t_idx]
            else:
                t_idx += 1

    def DeadCells(
        self,
        sequence_length: int,
        /,
        *,
        with_time_point: bool = False,
        with_living_leaves: bool = False,
        per_thread: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_thread_cell_h]:
        """"""
        if per_thread:
            output = {}
            AddDead = output.update
        else:
            output = []
            AddDead = output.extend

        for track in self:
            dead = track.DeadCells(
                sequence_length,
                with_time_point=with_time_point,
                with_living_leaves=with_living_leaves,
                per_thread=per_thread,
            )
            AddDead(dead)

        return output

    def DeathTimePoints(
        self,
        sequence_length: int,
        /,
        *,
        with_living_leaves: bool = False,
    ) -> Dict[int, Optional[int]]:
        """"""
        output = {}

        for track in self:
            time_points = track.DeathTimePoints(
                sequence_length,
                with_living_leaves=with_living_leaves,
            )
            output.update(time_points)

        return output

    def LeafCells(
        self, /, *, with_time_point: bool = False
    ) -> Union[cells_w_optional_time_points_h, per_thread_cells_h]:
        """"""
        leaves = []
        time_points = []
        for track in self:
            leaves.extend(track.leaves)
            time_points.extend(track.leaves_time_points)

        if with_time_point:
            return tuple(zip(leaves, time_points))

        return tuple(leaves)

    @property
    def total_n_topologic_threads(self) -> int:
        """
        total=among all structured tracks, not only valid ones
        """
        output = self.topologic_labels.__len__()

        for track in self.fully_pruned:
            output += track.topologic_labels.__len__()
        for track, _ in self.invalids:
            if isinstance(track, structured_track_t):
                output += track.topologic_labels.__len__()

        return output

    @property
    def all_structured_iterator(self) -> Iterator[structured_track_t]:
        """"""
        for track in self:
            yield track
        for track in self.fully_pruned:
            yield track
        for track, _ in self.invalids:
            if isinstance(track, structured_track_t):
                yield track

    @property
    def topologic_labels(self) -> Sequence[int]:
        """"""
        output = []

        for track in self:
            output.extend(track.topologic_labels)

        return sorted(output)

    @property
    def labels(self) -> Sequence[int]:
        """"""
        output = []

        for track in self:
            output.extend(track.labels)

        return sorted(output)

    def TrackLabelsContainingCell(self, cell: cell_t, /) -> Sequence[int]:
        """"""
        for track in self:
            if cell in track:
                return track.TrackLabelsContainingCell(cell)

        raise ValueError(f"{cell}: Not a tracked cell")

    def AddFeature(
        self,
        name: str,
        cell_feature_name: Union[str, Sequence[str]],
        FeatureComputation: feature_filtering_h,
        /,
        *args,
        **kwargs,
    ) -> None:
        """
        /!\ There is no already-existing check
        """
        if isinstance(cell_feature_name, str):
            cell_feature_names = (cell_feature_name,)
        else:
            cell_feature_names = cell_feature_name
        available_features = self[0].topologic_root.available_features
        if any(_ftr not in available_features for _ftr in cell_feature_names):
            raise ValueError(f"{cell_feature_name}: Invalid cell feature(s)")

        for track in self:
            track.AddFeature(
                name, cell_feature_names, FeatureComputation, *args, **kwargs
            )

    def Feature(self, feature: str, /) -> per_thread_feature_h:
        """"""
        output = {}

        for track in self:
            output.update(track.features[feature])

        return output

    def Print(self) -> None:
        """"""
        for track in self:
            print(track)

    def PrintValidInvalidSummary(self, /, *, logger: logger_t = None) -> None:
        """"""
        n_invalids = []
        issues_per_type = {}
        for track_type in (unstructured_track_t, thread_track_t, forking_track_t):
            invalids = [
                _rcd for _rcd in self.invalids if isinstance(_rcd[0], track_type)
            ]
            track_type = track_type.__name__[:-2].replace("_", " ").capitalize()
            if invalids.__len__() == 0:
                number = f"{track_type}: None"
            else:
                number = f"{track_type}: {invalids.__len__()}"

                issues = (
                    f"    Track {_ThreadLabelsAsStr(_tck)}{isse.ISSUE_SEPARATOR}"
                    f"{isse.FactorizedIssuesAsStr(_iss, max_n_contexts=5)}"
                    for _tck, _iss in invalids
                )
                issues_per_type[track_type] = "\n".join(issues)

            n_invalids.append(number)

        n_invalids = ", ".join(n_invalids)
        issues_per_type = "\n".join(
            f"{_typ}:\n{_iss}" for _typ, _iss in issues_per_type.items()
        )

        message = (
            f"Tracks: valid={self.__len__()}; "
            f"invalid={n_invalids}; "
            f"fully pruned={self.fully_pruned.__len__()}\n"
            f"{issues_per_type}"
        )
        if logger is None:
            print(message)
        else:
            logger.info(message)

    def __hash__(self) -> int:
        """"""
        return hash(tuple(hash(_tck) for _tck in self))

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        return f"{self.__repr__()}:\n    {self.__len__()=}\n    {self.invalids=}\n    {self.fully_pruned=}"

    def __contains__(self, cell: cell_t, /) -> bool:
        """"""
        return any(cell in _tck for _tck in self)

    def __eq__(self, other) -> bool:
        """"""
        return jsnr.JsonStringOf(self) == jsnr.JsonStringOf(other)


def _TreeIncompatibleCells(track: unstructured_track_t, /) -> Optional[Sequence[str]]:
    """"""
    output = []

    for cell in track.nodes:
        if (n_predecessors := track.in_degree(cell)) > 1:
            output.append(
                f"Cell {track.nodes[cell][TIME_POINT]}.{cell.label}{isse.ISSUE_SEPARATOR}{n_predecessors} "
                f"predecessors. Expected=0 or 1."
            )

    if output.__len__() == 0:
        output = None

    return output


def _ThreadLabelsAsStr(track: any_track_h, /) -> str:
    """"""
    return "+".join(str(_lbl) for _lbl in sorted(track.labels))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from io import BytesIO as io_bytes_t
from pathlib import Path as path_t
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import json_any.json_any as jsnr
import numpy as nmpy
# import scipy.ndimage as imge
from skimage.transform import resize as Resize

from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.number import MAX_INT
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.acquisition.frame import AsCompressedArray, all_versions_h


array_t = nmpy.ndarray
channel_computation_h = Callable[[Dict[str, array_t], Dict[str, Any]], array_t]
# ...=[array_t, **kwargs]
transform_h = Callable[..., array_t]


class frames_t(List[io_bytes_t]):
    """"""

    @classmethod
    def __NewFromJsonDescription__(cls, description: Any, /) -> frames_t:
        """"""
        instance = cls()
        # Use extend (instead of cls(description)) to allow for frame compression
        instance.extend(description)

        return instance

    def __DescriptionForJSON__(self) -> Any:
        """"""
        return list(self)

    def append(self, frame: array_t) -> None:
        """"""
        list.append(self, AsCompressedArray(frame))

    def extend(self, frames: Sequence[array_t]) -> None:
        """"""
        for frame in frames:
            list.append(self, AsCompressedArray(frame))

    def __setitem__(self, index: int, frame: array_t) -> None:
        """"""
        list.__setitem__(self, index, AsCompressedArray(frame))

    def __getitem__(self, index: int) -> array_t:
        """"""
        frame = list.__getitem__(self, index)
        frame.seek(0)

        return nmpy.load(frame)["array"]

    def __iter__(self) -> Iterator[array_t]:
        """"""
        for frame in list.__iter__(self):
            frame.seek(0)
            yield nmpy.load(frame)["array"]


@dtcl.dataclass(repr=False, eq=False)
class sequence_t(Dict[str, frames_t]):
    """"""

    path: path_t = None
    shape: Tuple[int, int] = None
    original_length: int = None
    first_frame: int = None
    length: int = None
    base_channels: Sequence[str] = None
    main_channel: str = None

    NewFromJsonString: ClassVar[Callable[[str], sequence_t]] = jsnr.ObjectFromJsonString

    @classmethod
    def NewFromFrames(
        cls,
        frames: array_t,
        in_channel_names: Sequence[Optional[str]],
        main_channel: str,
        path: path_t,
        /,
        *,
        first_frame: int = 0,
        last_frame: int = MAX_INT,
        expected_shape: Sequence[int] = None,
    ) -> sequence_t:
        """
        in_channel_names: names equal to None or "___" or "---" indicate channels that should be discarded
        """
        # TODO: make this function accept various input shapes thanks to an additional arrangement parameter of the
        #     form THRC, T=time, H=channel, RC= row column. This requires that SequenceFromPath deals with TH combined
        #     dimension.
        if (n_dims := frames.ndim) not in (3, 4):
            raise ValueError(
                f"{n_dims}: Invalid number of dimensions of sequence with shape {frames.shape}. "
                f"Expected=3 or 4=(TIME POINTS*CHANNELS)xROWSxCOLUMNS or "
                f"TIME POINTSxCHANNELSxROWSxCOLUMNS."
            )
        n_in_channels = in_channel_names.__len__()

        if n_dims == 3:
            frames = frames[
                (first_frame * n_in_channels) : ((last_frame + 1) * n_in_channels), ...
            ]
        else:
            frames = frames[first_frame : (last_frame + 1), ...]
        first_frame = 0
        last_frame = MAX_INT

        if expected_shape is not None:
            if n_dims == 3:
                order_in = (0, 1, 2)
                order_out = (2, 0, 1)
                fixed_size = frames.shape[:1]
                to_be_resized = frames.shape[1:]
            else:
                order_in = (0, 1, 2, 3)
                order_out = (2, 3, 0, 1)
                fixed_size = frames.shape[:2]
                to_be_resized = frames.shape[2:]
            if to_be_resized != expected_shape:
                LOGGER.warning(
                    f"Resizing sequence from actual size {to_be_resized} (full shape={frames.shape}) "
                    f"to expected size {expected_shape}"
                )
                frames = nmpy.moveaxis(frames, order_in, order_out)
                frames = Resize(
                    frames, (*expected_shape, *fixed_size), preserve_range=True
                )
                frames = nmpy.moveaxis(frames, order_out, order_in)
                LOGGER.info(f"Resizing done. New sequence shape={frames.shape}")

        frames_of_channel = {}
        for name in in_channel_names:
            if (name is not None) and (name != "___") and (name != "---"):
                frames_of_channel[name] = frames_t()
        base_channel_names = tuple(frames_of_channel.keys())

        if n_dims == 3:
            c_idx = n_in_channels - 1
            time_point = -1
            for frame in frames:
                c_idx += 1
                if c_idx == n_in_channels:
                    c_idx = 0
                    time_point += 1

                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                name = in_channel_names[c_idx]
                if name in base_channel_names:
                    frames_of_channel[name].append(frame)
        else:
            for time_point, frame in enumerate(frames):
                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                for c_idx, channel in enumerate(frame):
                    name = in_channel_names[c_idx]
                    if name in base_channel_names:
                        frames_of_channel[name].append(channel)

        frames_of_base_channel = frames_of_channel[base_channel_names[0]]
        shape = frames_of_base_channel[0].shape
        length = frames_of_base_channel.__len__()
        instance = cls(
            path=path,
            shape=shape,
            original_length=frames.__len__(),
            first_frame=first_frame,
            length=length,
            base_channels=base_channel_names,
            main_channel=main_channel,
        )
        instance.update(frames_of_channel)

        return instance

    @classmethod
    def __NewFromJsonDescription__(
        cls,
        description: Tuple[Dict[str, frames_t], Dict[str, Any]],
        /,
    ) -> sequence_t:
        """"""
        frames, attributes = description

        instance = cls(**attributes)
        instance.update(frames)

        return instance

    def __DescriptionForJSON__(self) -> Any:
        """"""
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose information.
        attributes = {_fld.name: getattr(self, _fld.name) for _fld in dtcl.fields(self)}

        return dict(self), attributes

    def AddChannel(
        self, name: str, ChannelComputation: channel_computation_h, /, **kwargs
    ) -> None:
        """"""
        if name in self:
            raise ValueError(f"{name}: Existing channel cannot be overridden")

        computed = frames_t()
        for frames in self.FramesIterator(channels=self.channels):
            frames_as_dict = dict(zip(self.channels, frames))
            frame = ChannelComputation(frames_as_dict, **kwargs)
            computed.append(frame)

        self[name] = computed

    @property
    def channels(self) -> Tuple[str, ...]:
        """
        Names of channels read from file (base channels) and computed channels
        """
        return tuple(self.keys())

    def ChannelExtrema(self, channel: str, /) -> Tuple[float, float]:
        """"""
        min_intensity = nmpy.Inf
        max_intensity = -nmpy.Inf

        for frame in self[channel]:
            min_intensity = min(min_intensity, nmpy.amin(frame))
            max_intensity = max(max_intensity, nmpy.amax(frame))

        # Apparently, the intensities might not be Numpy object in certain situations (Numpy version?). Although
        # currently unexplained, an easy workaround is to return the intensities as is.
        try:
            output = min_intensity.item(), max_intensity.item()
        except AttributeError:
            output = min_intensity, max_intensity

        return output

    def FramesIterator(
        self,
        /,
        *,
        channels: Sequence[str] = None,
    ) -> Iterator[Sequence[array_t]]:
        """
        channel: None=all (!) base channels
        """
        if channels is None:
            channels = self.base_channels

        for f_idx in range(self.length):
            output = tuple(self[_chl][f_idx] for _chl in channels)
            yield output

    def ApplyTransform(
        self,
        Transform: transform_h,
        /,
        *,
        channel: Union[str, Sequence[str]] = None,
        **kwargs,
    ) -> None:
        """
        channel: None=all (!) base channels
        """
        if channel is None:
            channels = self.base_channels
        elif isinstance(channel, str):
            channels = (channel,)
        else:
            channels = channel

        for channel in channels:
            targets = self[channel]
            references_sets = self.FramesIterator(channels=self.channels)
            for target, references in zip(targets, references_sets):
                refs_as_dict = dict(zip(self.channels, references))
                transformed = Transform(target, channels=refs_as_dict, **kwargs)
                target[...] = transformed[...]

    def __hash__(self) -> int:
        """"""
        return hash((self.path, self.shape, self.first_frame, self.length))

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        all_extrema = []
        for channel in self.channels:
            extrema = self.ChannelExtrema(channel)
            all_extrema.append(f"{channel}=[{extrema[0]},{extrema[1]}]")
        all_extrema = "\n    ".join(all_extrema)

        return (
            f"{repr(self)}:\n"
            f"    {self.path=}\n"
            f"    {self.base_channels=}\n"
            f"    {self.channels=}\n"
            f"    {all_extrema}\n"
            f"    {self.shape=}\n"
            f"    {self.original_length=}\n"
            f"    {self.length=}\n"
        )

    def __eq__(self, other, /) -> bool:
        """"""
        return self.DifferencesWith(other).__len__() == 0

    def DifferencesWith(self, other, /) -> Sequence[str]:
        """"""
        if type(other) is type(self):
            output = []

            for name, frames in self.items():
                other_frames = other[name]
                if frames.__len__() != other_frames.__len__():
                    output.append(
                        f"Lengths {name}: {frames.__len__()} != {other_frames.__len__()}"
                    )
                for f_idx, (frame, other_frame) in enumerate(zip(frames, other_frames)):
                    if not nmpy.array_equal(frame, other_frame):
                        output.append(
                            f"{name}[{f_idx}]: ({frame.shape}, {frame.dtype}) != ({other_frame.shape}, {other_frame.dtype})"
                        )

            for field in dtcl.fields(self):
                name = field.name
                value = getattr(self, name)
                other_value = getattr(other, name)
                if value is None:
                    if other_value is not None:
                        output.append(f"{name}: None != {other_value}")
                elif other_value is None:
                    output.append(f"{name}: {value} != None")
                elif other_value != value:
                    output.append(f"{name}: {value} != {other_value}")

            return output

        return (f"Type: {type(self).__name__} != {type(other).__name__}",)


# def BoundingBoxSlices(sequence: array_t, /) -> Tuple[slice, slice]:
#     """
#     sequence: as an XYCT-volume
#     """
#     min_reduction = nmpy.amin(sequence, axis=-1)
#     max_reduction = nmpy.amax(sequence, axis=-1)
#     variable = nmpy.any(min_reduction != max_reduction, axis=-1)
#
#     output: Sequence = imge.find_objects(variable.astype(nmpy.int8))
#     if output.__len__() == 0:
#         raise ValueError(f"{sequence}: Empty sequence")
#
#     return output[0]


def AllChannels(
    sequence: Union[Sequence[array_t], sequence_t], /
) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, sequence_t):
        all_channels = {}
        for channel, frames in sequence.items():
            min_value = min(nmpy.amin(_frm) for _frm in frames)
            max_value = max(nmpy.amax(_frm) for _frm in frames)
            all_channels[channel] = ((min_value, max_value), frames)
        current_channel = sequence.channels[0]
    else:
        current_channel = "MAIN"
        min_value = min(nmpy.amin(_frm) for _frm in sequence)
        max_value = max(nmpy.amax(_frm) for _frm in sequence)
        all_channels = {current_channel: ((min_value, max_value), sequence)}

    return all_channels, current_channel



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numbers as nmbr
import tempfile as temp
from pathlib import Path as path_t
from typing import IO, Dict, List, Optional, Sequence, Tuple, Union

import xlsxwriter as xlsx
from xlsxwriter.format import Format as xlsx_format_t
from xlsxwriter.workbook import Workbook as workbook_t
from xlsxwriter.worksheet import Worksheet as worksheet_t

from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.uid import AlphaColumnFromIndex
from cell_tracking_BC.type.analysis import analysis_t
from cell_tracking_BC.type.cell.cell import cell_t, state_e
from cell_tracking_BC.type.track.single.structured import structured_track_t
from cell_tracking_BC.type.track.single.thread import thread_track_t
from cell_tracking_BC.type.track.multiple.structured import tracks_t


division_times_h = Dict[int, Sequence[int]]
death_time_h = Dict[int, int]


INVALID_TRACK_MARKER = "Invalid"
PRUNED_TRACK_MARKER = "Pruned"
VALID_TRACK_MARKER = "Valid"

_DASH_TYPES = (
    "solid",
    "round_dot",
    "square_dot",
    "dash",
    "dash_dot",
    "long_dash",
    "long_dash_dot",
    "long_dash_dot_dot",
)
_N_DASH_TYPES = _DASH_TYPES.__len__()


def SaveCellFeaturesToXLSX(
    path: Union[str, path_t],
    analysis: analysis_t,
    /,
    *,
    feature: Union[str, Sequence[str]] = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if feature is None:
        features = analysis.segmentations.available_cell_features
    elif isinstance(feature, str):
        features = (feature,)
    else:
        features = feature

    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))
    pruned_format = workbook.add_format({"bg_color": "gray"})
    division_format = workbook.add_format({"bg_color": "blue"})
    death_format = workbook.add_format({"bg_color": "red"})

    csv_path = path.with_suffix(".csv")
    csv_accessor = open(csv_path, mode="w")

    tracks = analysis.tracks
    if (tracks is None) or (tracks.__len__() == 0):
        LOGGER.warning("Sequence with no valid tracks.")
    else:
        _SaveSiblings(tracks, workbook, csv_accessor)
        _SaveTrackStatus(tracks, workbook, csv_accessor)

        for feature in features:
            _SaveCellFeature(
                feature,
                tracks,
                workbook,
                csv_accessor,
                pruned_format,
                division_format,
                death_format,
            )

    workbook.close()
    csv_accessor.close()


def _SaveSiblings(
    tracks: tracks_t,
    workbook: workbook_t,
    csv_accessor: IO,
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName("siblings")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Siblings\n")

    csv_lines = []
    for track in tracks.all_structured_iterator:
        for c_idx, (labels, label_type) in enumerate(
            zip(
                (set(track.labels), set(track.topologic_labels)),
                ("", "Topologic "),
            )
        ):
            for label in labels:
                other_labels = str(sorted(labels.difference((label,))))[1:-1]
                if other_labels.__len__() > 0:
                    worksheet.write_string(label - 1, c_idx, other_labels)
                    csv_lines.append(f"{label}, {label_type}Labels: {other_labels}")
                else:
                    if label_type.__len__() > 0:
                        message = f"Thread track in {label_type}mode"
                    else:
                        message = "Thread track"
                    worksheet.write_string(label - 1, c_idx, message)
                    csv_lines.append(f"{label}, {message}")

    worksheet.write_string(tracks.total_n_topologic_threads, 0, "END")
    _SortAndWriteCSVLines(csv_lines, csv_accessor)


def _SaveTrackStatus(
    tracks: tracks_t,
    workbook: workbook_t,
    csv_accessor: IO,
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName("status")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Status\n")

    csv_lines = []
    invalids = (
        _rcd[0] for _rcd in tracks.invalids if isinstance(_rcd[0], structured_track_t)
    )
    for tracks_per_status, status in zip(
        (tracks, tracks.fully_pruned, invalids),
        (VALID_TRACK_MARKER, PRUNED_TRACK_MARKER, INVALID_TRACK_MARKER),
    ):
        for track in tracks_per_status:
            for label in track.topologic_labels:
                if (label in track.labels) or (status != VALID_TRACK_MARKER):
                    thread_status = status
                else:
                    thread_status = PRUNED_TRACK_MARKER
                worksheet.write_string(label - 1, 0, thread_status)
                csv_lines.append(f"{label}, {thread_status}")

    worksheet.write_string(tracks.total_n_topologic_threads, 0, "END")
    _SortAndWriteCSVLines(csv_lines, csv_accessor)


def _SaveCellFeature(
    feature: str,
    tracks: tracks_t,
    workbook: workbook_t,
    csv_accessor: IO,
    pruned_format: Optional[xlsx_format_t],
    division_format: Optional[xlsx_format_t],
    death_format: Optional[xlsx_format_t],
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName(feature)
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write(f"--- Feature {feature}\n")

    per_row_limits = {}
    csv_lines = []
    for track in tracks.all_structured_iterator:
        root_time_point = track.topologic_root_time_point
        for path, label in track.LabeledThreadIterator(topologic_mode=True):
            row = label - 1

            if feature not in path[0].features:
                message = f'Feature "{feature}" missing'
                worksheet.write_string(row, 0, message)
                csv_lines.append(f"{label}, {message}")
                continue

            evolution = tuple(_cll.features[feature] for _cll in path)
            if not isinstance(evolution[0], nmbr.Number):
                message = f'Feature of type "{type(evolution[0]).__name__}" unhandled'
                worksheet.write_string(0, 0, message)
                csv_accessor.write(message + "\n")
                return

            if root_time_point > 0:
                worksheet.write_row(row, 0, root_time_point * ("x",))
            worksheet.write_row(row, root_time_point, evolution)
            csv_lines.append(
                f"{label}, " + root_time_point * "," + ", ".join(map(str, evolution))
            )

            _SetStateBasedCellFormat(
                worksheet,
                row,
                root_time_point,
                path,
                evolution,
                pruned_format,
                division_format,
                death_format,
            )

            per_row_limits[row + 1] = (
                root_time_point,
                root_time_point + evolution.__len__() - 1,
            )

    _WriteExtras(
        tracks,
        per_row_limits,
        workbook,
        sheet_name,
        worksheet,
        pruned_format,
        division_format,
        death_format,
    )
    _SortAndWriteCSVLines(csv_lines, csv_accessor)


def SaveCellEventsToXLSX(
    path: Union[str, path_t],
    analysis: analysis_t,
    /,
    *,
    division_response: str = None,
    death_response: str = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))
    pruned_format = workbook.add_format({"bg_color": "gray"})
    division_format = workbook.add_format({"bg_color": "blue"})
    death_format = workbook.add_format({"bg_color": "red"})

    csv_path = path.with_suffix(".csv")
    csv_accessor = open(csv_path, mode="w")

    tracks = analysis.tracks
    if (tracks is None) or (tracks.__len__() == 0):
        LOGGER.warning("Sequence with no valid tracks.")
    else:
        _SaveSiblings(tracks, workbook, csv_accessor)
        _SaveTrackStatus(tracks, workbook, csv_accessor)

        division_time_points = tracks.DivisionTimePoints()
        death_time_points = tracks.DeathTimePoints(
            analysis.sequence.length, with_living_leaves=True
        )
        for SaveEvents, time_point in zip(
            (_SaveDivisionEvents, _SaveDeathEvents),
            (division_time_points, death_time_points),
        ):
            SaveEvents(workbook, csv_accessor, time_point, tracks)

        n_divisions = sum(_tck.n_dividing_cells for _tck in tracks)
        _SaveEventCounts(workbook, csv_accessor, n_divisions, death_time_points)

        for event, response in zip(
            ("division", "death"), (division_response, death_response)
        ):
            if response is not None:
                _SaveEventResponse(
                    event,
                    workbook,
                    csv_accessor,
                    response,
                    tracks,
                    pruned_format,
                    division_format,
                    death_format,
                )

    workbook.close()
    csv_accessor.close()


def _SaveDivisionEvents(
    workbook: xlsx.Workbook,
    csv_accessor: IO,
    division_time_points: Optional[division_times_h],
    tracks: tracks_t,
    /,
) -> None:
    """"""
    if (division_time_points is None) or (division_time_points.__len__() == 0):
        LOGGER.warning("Division Events: No associated tracks.")
        return

    sheet_name = _SheetNameFromLongName("division times")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Division Times\n")

    total_n_topologic_threads = tracks.total_n_topologic_threads
    for label in range(1, total_n_topologic_threads + 1):
        row = label - 1

        if label not in division_time_points:
            worksheet.write_string(
                row, 0, f"{INVALID_TRACK_MARKER} or {PRUNED_TRACK_MARKER}"
            )
            csv_accessor.write(
                f"{label}:, {INVALID_TRACK_MARKER} or {PRUNED_TRACK_MARKER}\n"
            )
            continue

        time_points = division_time_points[label]
        if (time_points is None) or (time_points.__len__() == 0):
            worksheet.write_string(row, 0, "No Divisions")
            csv_accessor.write(f"{label}:, No Divisions\n")
        else:
            worksheet.write_row(row, 0, time_points)
            csv_accessor.write(f"{label}:," + ", ".join(map(str, time_points)) + "\n")

    worksheet.write_string(total_n_topologic_threads, 0, "END")


def _SaveDeathEvents(
    workbook: xlsx.Workbook,
    csv_accessor: IO,
    death_time_points: Optional[death_time_h],
    tracks: tracks_t,
    /,
) -> None:
    """"""
    if (death_time_points is None) or (death_time_points.__len__() == 0):
        LOGGER.warning("Death Events: No associated tracks.")
        return

    sheet_name = _SheetNameFromLongName("death time")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Death Times\n")

    total_n_topologic_threads = tracks.total_n_topologic_threads
    for label in range(1, total_n_topologic_threads + 1):
        row = label - 1

        if label not in death_time_points:
            worksheet.write_string(
                row, 0, f"{INVALID_TRACK_MARKER} or {PRUNED_TRACK_MARKER}"
            )
            csv_accessor.write(
                f"{label}:, {INVALID_TRACK_MARKER} or {PRUNED_TRACK_MARKER}\n"
            )
            continue

        time_point = death_time_points[label]
        if time_point is None:
            worksheet.write_string(row, 0, "No Death")
            csv_accessor.write(f"{label}:, No Death\n")
        else:
            worksheet.write_number(row, 0, time_point)
            csv_accessor.write(f"{label}:, {time_point}\n")

    worksheet.write_string(total_n_topologic_threads, 0, "END")


def _SaveEventCounts(
    workbook: xlsx.Workbook,
    csv_accessor: IO,
    n_divisions: int,
    death_time_points: Optional[death_time_h],
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName("event counts")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write("--- Event Counts\n")

    n_deaths_pattern = 0
    n_deaths_track = 0
    if death_time_points is not None:
        for time_point in death_time_points.values():
            if time_point is not None:
                if time_point >= 0:
                    n_deaths_pattern += 1
                else:
                    n_deaths_track += 1

    for r_idx, (title, value) in enumerate(
        zip(
            ("divisions", "death (pattern)", "death (topologic)", "death"),
            (
                n_divisions,
                n_deaths_pattern,
                n_deaths_track,
                n_deaths_pattern + n_deaths_track,
            ),
        )
    ):
        worksheet.write_string(r_idx, 0, title)
        worksheet.write_number(r_idx, 1, value)
        csv_accessor.write(f"{title}, {value}\n")


def _SaveEventResponse(
    event: str,
    workbook: xlsx.Workbook,
    csv_accessor: IO,
    name: str,
    tracks: tracks_t,
    pruned_format: xlsx_format_t,
    division_format: xlsx_format_t,
    death_format: xlsx_format_t,
    /,
) -> None:
    """"""
    sheet_name = _SheetNameFromLongName(f"{event} response")
    worksheet = workbook.add_worksheet(sheet_name)
    csv_accessor.write(f"--- {event.title()} response\n")

    per_row_limits = {}
    csv_lines = []
    for track in tracks.all_structured_iterator:
        if name not in track.features:
            message = f'Track without "{name}" response'
            for label in track.topologic_labels:
                worksheet.write_string(label - 1, 0, message)
                csv_lines.append(f"{label}, {message}")
            continue

        root_time_point = track.topologic_root_time_point
        for label, response in track.features[name].items():
            row = label - 1
            if response is None:
                message = "Track too Short for Valid Response"
                worksheet.write_string(row, 0, message)
                csv_lines.append(f"{label}, {message}")
            else:
                worksheet.write_row(row, root_time_point, response)
                csv_lines.append(
                    f"{label}, " + root_time_point * "," + ", ".join(map(str, response))
                )

                if isinstance(track, thread_track_t):
                    cells = track
                else:
                    leaf_idx = track.topologic_labels.index(label)
                    leaf = track.topologic_leaves[leaf_idx]
                    cells = track.PathFromTo(track.topologic_root, leaf)
                _SetStateBasedCellFormat(
                    worksheet,
                    row,
                    root_time_point,
                    cells,
                    response,
                    pruned_format,
                    division_format,
                    death_format,
                )

                per_row_limits[label] = (
                    root_time_point,
                    root_time_point + response.__len__() - 1,
                )

    _WriteExtras(
        tracks,
        per_row_limits,
        workbook,
        sheet_name,
        worksheet,
        pruned_format,
        division_format,
        death_format,
    )
    _SortAndWriteCSVLines(csv_lines, csv_accessor)


def _SheetNameFromLongName(name: str, /) -> str:
    """
    Sheet names cannot exceed 31 characters in length:
    xlsxwriter.exceptions.InvalidWorksheetName: Excel worksheet name 'cfp-background_over_yfp-background' must be <= 31 chars.
    """
    LIMIT = 31

    if name.__len__() <= LIMIT:
        return name

    return f"{name[:(LIMIT-3)]}..."


def _WriteExtras(
    tracks: tracks_t,
    per_row_limits: Dict[int, Tuple[int, int]],
    workbook: workbook_t,
    sheet_name: str,
    worksheet: worksheet_t,
    pruned_format: xlsx_format_t,
    division_format: xlsx_format_t,
    death_format: xlsx_format_t,
    /,
) -> None:
    """"""
    next_available_row = tracks.total_n_topologic_threads

    worksheet.write_string(next_available_row, 0, "END")
    next_available_row += 2  # With margin

    _AddCellStateLegend(
        worksheet, next_available_row, pruned_format, division_format, death_format
    )
    next_available_row += 3  # With margin

    if per_row_limits.__len__() > 0:
        chart = workbook.add_chart({"type": "line"})
        for l_idx, (row, (min_col, max_col)) in enumerate(per_row_limits.items()):
            min_col = AlphaColumnFromIndex(min_col)
            max_col = AlphaColumnFromIndex(max_col)
            chart.add_series(
                {
                    "name": str(row),
                    "values": f"='{sheet_name}'!${min_col}${row}:${max_col}${row}",
                    "line": {
                        "width": 1.0,
                        "dash_type": _DASH_TYPES[l_idx % _N_DASH_TYPES],
                    },
                }
            )
        worksheet.insert_chart(f"A{next_available_row}", chart)


def _SetStateBasedCellFormat(
    worksheet: worksheet_t,
    row: int,
    root_time_point: int,
    cells: Sequence[cell_t],
    values: Sequence,
    pruned_format: xlsx_format_t,
    division_format: xlsx_format_t,
    death_format: xlsx_format_t,
    /,
) -> None:
    """"""
    for time_point, cell in enumerate(cells, start=root_time_point):
        if cell.state is state_e.discarded:
            cell_format = pruned_format
        elif cell.state is state_e.dividing:
            cell_format = division_format
        elif cell.state is state_e.dead:
            cell_format = death_format
        else:
            cell_format = None

        if cell_format is not None:
            value = values[time_point - root_time_point]
            worksheet.write_number(row, time_point, value, cell_format)


def _AddCellStateLegend(
    worksheet: worksheet_t,
    row: int,
    pruned_format: Optional[xlsx_format_t],
    division_format: Optional[xlsx_format_t],
    death_format: Optional[xlsx_format_t],
    /,
) -> None:
    """"""
    for col, state, cell_format in zip(
        range(3),
        ("Dividing", "Dead", "Pruned"),
        (division_format, death_format, pruned_format),
    ):
        if cell_format is not None:
            worksheet.write_string(row, col, state, cell_format)


def _SortAndWriteCSVLines(csv_lines: List[str], csv_accessor: IO, /) -> None:
    """"""
    csv_lines.sort(key=lambda _lne: int(_lne.split(",", maxsplit=1)[0]))
    csv_accessor.writelines(_lne + "\n" for _lne in csv_lines)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import bz2 as bzp2
import sys as sstm
from pathlib import Path as path_t
from typing import Any, Optional, Union

import json_any.json_any as jsny
from json_any.test.test_json_any import DecodeAndCompare

import cell_tracking_BC.in_out.file.path as path
from cell_tracking_BC.in_out.text.logger import LOGGER


def Save(
    element: Any, name: str, folder: path_t, check_saving: bool, /, *, builders=None
) -> None:
    """"""
    full_name = folder / f"{name}.json"
    if full_name.exists():
        raise ValueError(f"{name}: JSON file already exists; Exiting")

    jsoned = jsny.JsonStringOf(element)
    compressed = bzp2.compress(jsoned.encode(), compresslevel=2)
    with open(full_name, "bw") as accessor:
        accessor.write(compressed)

    if check_saving:
        if hasattr(type(element), "NewFromJsonString") and hasattr(
            element, "DifferencesWith"
        ):
            decoded = type(element).NewFromJsonString(jsoned)
            differences = element.DifferencesWith(decoded)
            if differences.__len__() > 0:
                LOGGER.error("\n".join(differences))
                sstm.exit(1)
        elif not DecodeAndCompare(jsoned, element, builders=builders):
            sstm.exit(1)


def Load(
    type_: Any, name: str, folder: path_t, /, *, builders=None
) -> Optional[object]:
    """"""
    full_name = folder / f"{name}.json"
    if not full_name.is_file():
        return None

    with open(full_name, "br") as accessor:
        compressed = accessor.read()

    jsoned = bzp2.decompress(compressed).decode()

    if hasattr(type_, "NewFromJsonString"):
        return type_.NewFromJsonString(jsoned)

    return jsny.ObjectFromJsonString(jsoned, builders=builders)


def ReplayFolderForMain(
    main: Union[str, path_t],
    folder_name: str,
    /,
    *,
    base_folder_name: str = "_replay",
    base_folder_postfix: str = "",
    play_date: str = None,
) -> path_t:
    """
    Call with main equal to __file__
    play_date: Can be None (play mode), or (replay mode) "latest" or a date-time in iso format with
        timespec='milliseconds'
    """
    if isinstance(main, str):
        main = path_t(main)

    output = main.parent / (base_folder_name + base_folder_postfix)
    output = path.ReplacePathIllegalCharacters(output)
    if not output.exists():
        # exist_ok: added for thread safety (could use only mkdir then...)
        output.mkdir(exist_ok=True)
    elif not output.is_dir():
        LOGGER.error(f"{output}: Not a suitable replay folder")
        sstm.exit(-1)

    output /= path.ReplacePathIllegalCharacters(folder_name)
    if not output.exists():
        # exist_ok: added for thread safety (could use only mkdir then...)
        output.mkdir(exist_ok=True)
    elif not output.is_dir():
        LOGGER.error(f"{output}: Not a suitable replay folder")
        sstm.exit(-1)

    if play_date is None:
        output /= path.TimeStamp()
        if output.exists():
            LOGGER.error(f"{output}: Existing date-based replay folder; Exiting")
            sstm.exit(-1)
        output.mkdir()
    elif play_date == "latest":
        everything = sorted(output.glob("*"))
        output /= everything[-1]
    else:
        output /= play_date
    if play_date is not None:
        if not output.is_dir():
            LOGGER.error(f"{output}: Not a valid replay folder")
            sstm.exit(-1)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.type.cell.cell import cell_t, state_e
from cell_tracking_BC.type.track.single.structured import (
    cells_w_optional_time_points_h,
    per_thread_cells_h,
    structured_track_t,
)


# TODO: check if grph.descendants and grph.ancestors could be used in place of more complex code here
# TODO: cell iteration methods should have a topologic_mode: bool = False parameter to allow cell-state-independent
#     iterations


class thread_track_t(structured_track_t, List[cell_t]):
    label: int = None
    tracking_affinities: Tuple[float, ...] = None

    def __init__(
        self,
        cells: Sequence[cell_t],
        tracking_affinities: Sequence[float],
        root_time_point: int,
        label: Optional[int],
        /,
    ) -> None:
        """"""
        list.__init__(self, cells)

        # Place after list init to get access to self[...]
        structured_track_t.__init__(
            self,
            root=self[0],
            leaves=(self[-1],),
        )
        self.topologic_root_time_point = root_time_point
        self.topologic_labels = (label,)
        self._dividing_marked = True

        self.label = label
        self.tracking_affinities = tuple(tracking_affinities)

    @classmethod
    def NewFromOrderedCells(
        cls,
        cells: Sequence[cell_t],
        tracking_affinities: Sequence[float],
        root_time_point: int,
        label: Optional[int],
        /,
    ) -> thread_track_t:
        """
        This must be the only place where direct instantiation is allowed. Anywhere else, instantiation must be
        performed with this class method.

        label: Can be None only to accommodate the creation of branches as threads
        """
        n_cells = cells.__len__()
        if (n_affinities := tracking_affinities.__len__()) != (
            n_expected := n_cells - 1
        ):
            raise ValueError(
                f"{n_affinities}: Invalid number of tracking_affinities. Expected={n_expected}"
            )

        return cls(cells, tracking_affinities, root_time_point, label)

    @classmethod
    def __NewFromJsonDescription__(
        cls,
        description,
        /,
    ) -> thread_track_t:
        """"""
        cells, label, tracking_affinities, attributes = description

        instance = cls(
            cells, tracking_affinities, attributes["topologic_root_time_point"], label
        )
        for name, value in attributes.items():
            setattr(instance, name, value)

        return instance

    def __DescriptionForJSON__(self) -> Any:
        """"""
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose information.
        attributes = {
            _fld.name: getattr(self, _fld.name)
            for _fld in dtcl.fields(structured_track_t)
        }

        return list(self), self.label, self.tracking_affinities, attributes

    @property
    def unpruned_cells(self) -> Sequence[cell_t]:
        """"""
        return self[self.index(self.root) : (self.index(self.leaves[0]) + 1)]

    def CellTimePoint(self, cell: cell_t, /) -> int:
        """"""
        return self.topologic_root_time_point + self.index(cell)

    def SplittingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        return ()

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """"""
        LOGGER.warning(f"{self}: Dividing cells already marked")

    @property
    def n_dividing_cells(self) -> int:
        """"""
        return 0

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_thread: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_thread_cells_h]:
        """"""
        if per_thread:
            return {}
        return ()

    @property
    def labels(self) -> Sequence[int]:
        """
        This method is supposed to be called from not fully-pruned multiple.
        """
        return (self.label,)  # Note: this is a tuple

    @property
    def pruned_labels(self) -> Sequence[int]:
        """
        This method is supposed to be called from not fully-pruned multiple.
        """
        return ()

    def TrackLabelsContainingCell(self, cell: cell_t, /) -> Sequence[int]:
        """
        Returns valid sequence even if cell is not in self
        """
        if cell.state is state_e.discarded:
            return ()

        return self.labels

    def TrackLabelWithLeaf(self, _: cell_t, /) -> int:
        """
        Returns label even if cell is not in self or is not a leaf
        """
        return self.label

    def TrackingAffinitiesBetween(
        self, first: cell_t, second: cell_t, /
    ) -> Sequence[float]:
        """"""
        return self.tracking_affinities[self.index(first) : self.index(second)]

    def CellSuccessors(self, cell: cell_t, /) -> Sequence[cell_t]:
        """"""
        where = self.index(cell)
        if where < self.__len__() - 1:
            successor = self[where + 1]
            if successor.state is state_e.discarded:
                output = ()
            else:
                output = (successor,)
        else:
            output = ()

        return output

    def CellDescendants(
        self,
        cell: cell_t,
        /,
        *,
        including_self: bool = True,
        topologic_mode: bool = False,
    ) -> Sequence[cell_t]:
        """"""
        where = self.index(cell)
        if not including_self:
            where += 1

        if where > self.__len__() - 1:
            output = ()
        else:
            output = tuple(self[where:])
            if not topologic_mode:
                pruned = tuple(_cll.state is state_e.discarded for _cll in output)
                if any(pruned):
                    first_pruned = pruned.index(True)
                    output = output[:first_pruned]

        return output

    def ConfirmCellLineage(self, youngest: cell_t, oldest: cell_t, /) -> bool:
        """"""
        return self.index(youngest) <= self.index(oldest)

    def PathFromTo(self, first: cell_t, last: cell_t, /) -> Sequence[cell_t]:
        """"""
        return self[self.index(first) : (self.index(last) + 1)]

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        root_idx = self.index(self.root)
        leaf_idx = self.index(self.leaves[0])
        for c_idx in range(root_idx, leaf_idx):
            time_point = self.topologic_root_time_point + c_idx
            is_last = c_idx == leaf_idx - 1
            yield time_point, *self[c_idx : (c_idx + 2)], is_last

    def PiecesIterator(
        self, /, *, with_affinities: bool = False, topologic_mode: bool = False
    ) -> Iterator[
        Union[
            Tuple[Sequence[cell_t], int, Optional[int]],
            Tuple[Sequence[cell_t], int, Optional[int], Sequence[float]],
        ]
    ]:
        """
        int: time point of first cell
        Optional[int]: thread label if last cell is (topologic) leaf
        """
        if topologic_mode:
            if with_affinities:
                return iter(
                    (
                        (
                            self,
                            self.topologic_root_time_point,
                            self.label,
                            self.tracking_affinities,
                        ),
                    )
                )
            else:
                return iter(((self, self.topologic_root_time_point, self.label),))
        else:
            if with_affinities:
                return iter(
                    (
                        (
                            self.PathFromTo(self.root, self.leaves[0]),
                            self.root_time_point,
                            self.label,
                            self.TrackingAffinitiesBetween(self.root, self.leaves[0]),
                        ),
                    )
                )
            else:
                return iter(
                    (
                        (
                            self.PathFromTo(self.root, self.leaves[0]),
                            self.root_time_point,
                            self.label,
                        ),
                    )
                )

    def LabeledThreadIterator(
        self,
        /,
        *,
        topologic_mode: bool = False,
    ) -> Iterator[Tuple[Sequence[cell_t], int]]:
        """"""
        if topologic_mode:
            output = (self, self.label)
        else:
            output = (self.unpruned_cells, self.label)

        return iter((output,))

    def CellFeature(
        self, feature: str, /, *, topologic_mode: bool = False
    ) -> Sequence[Any]:
        """
        Contrary to structured_track_t.features which are rooted at time point zero using Nones if needed, here the
        feature is returned for the unpruned part only.
        """
        if topologic_mode:
            return tuple(_cll.features[feature] for _cll in self)

        return tuple(_cll.features[feature] for _cll in self.unpruned_cells)
